
local params = ...

local tooltip_rim = CreateFrame("Frame")
tooltip_rim:SetLayer(500)
tooltip_rim:SetPoint("origin", tooltip_rim, "TOPLEFT")

local tooltip_bg = CreateFrame("Frame", tooltip_rim)
tooltip_bg:SetPoint("TOPLEFT", "origin", 2, 2)

local tooltip = CreateFrame("Text_Multiline", tooltip_bg)
tooltip:SetWidth(280)
tooltip:SetSize(15)
tooltip:SetText("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum")
tooltip:SetColor(1, 1, 1)
tooltip:SetPoint("TOPLEFT", tooltip_bg, "TOPLEFT", 2, 2)

tooltip_bg:SetPoint("BOTTOMRIGHT", tooltip, "BOTTOMRIGHT", 2, 2)

tooltip_rim:SetHandle("x", "size", {tooltip_bg:GetHandle("x", "size")}, function (p) return p[1]:Get() + 4 end)
tooltip_rim:SetHandle("y", "size", {tooltip_bg:GetHandle("y", "size")}, function (p) return p[1]:Get() + 4 end)

tooltip_rim:Hide()

function tooltip:Tick()
  local mx, my = GetMouseX(), GetMouseY()
  local ttx, tty = tooltip_rim:GetWidth(), tooltip_rim:GetHeight()
  
  local clamp_up = false
  local clamp_left = false

  if ttx > mx + 5 then clamp_left = true end
  if tty > my + 5 then clamp_up = true end
  
  tooltip_rim:ClearPoints("alignment", "alignment")
  if clamp_up then
    tooltip_rim:SetPoint("TOP", "origin")
  else
    tooltip_rim:SetPoint("BOTTOM", "origin", nil, my - 5)
  end
  if clamp_left then
    tooltip_rim:SetPoint("LEFT", "origin")
  else
    tooltip_rim:SetPoint("RIGHT", "origin", mx - 5, nil)
  end
end

tooltip_bg:SetBackgroundColor(0, 0, 0.3)
tooltip_rim:SetBackgroundColor(0.4, 0.4, 1.0)

local game_ended = false

function tooltippy(frame, func)
  local omin, omout = frame.MouseIn, frame.MouseOut
  function frame:MouseIn()
    if omin then omin(self) end
    
    if game_ended then return end
    
    local tex = func()
    if tex then
      tooltip_rim:Show()
      tooltip:SetText(tex)
    end
  end
  function frame:MouseOut()
    if omout then omout(self) end
    
    tooltip_rim:Hide()
  end
end

local bg = CreateFrame("Texture")
bg:SetTexture("bg")
bg:ClearPoints("all", "all")
bg:SetAllPoints()
bg:SetColor(0.8, 0.8, 0.8)

local function endgame(message)
  local bgscreen = CreateFrame("Frame", bg)
  bgscreen:SetLayer(1000)
  bgscreen:SetBackgroundColor(0, 0, 0, 0.7)
  bgscreen:SetAllPoints()
  
  local texov = CreateFrame("Text", bgscreen)
  texov:SetColor(1, 1, 1)
  texov:SetText(message)
  texov:SetPoint("CENTER", bgscreen, "CENTER")
  texov:SetSize(80)
  
  endgame_freeze = true
  
  local cb = CreateFrame("Button", bgscreen)
  
  local cont = CreateFrame("Text", cb)
  cont:SetColor(1, 1, 1)
  cont:SetText("End game")
  cont:SetPoint("CENTER", bgscreen, "CENTER", 0, 300)
  
  cb:SetPoint("TOPLEFT", cont, "TOPLEFT", -40, -20)
  cb:SetPoint("BOTTOMRIGHT", cont, "BOTTOMRIGHT", 40, 20)
  cb:SetBackgroundColor(1, 1, 1, 0.2)
  
  function cb:Click()
    GlorpController("exit_game")
  end
  
  game_ended = true
end

local status = CreateFrame("Text", bg)
status:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", 3, -3)
status:SetColor(1, 1, 1)
status:SetSize(20)

local event
local inside_x, inside_y

local map = {}
local map_targeted_shown = false

for i = 1, 32 do
  local ro = {}
  for i = 1, 32 do
    table.insert(ro, {})
  end
  table.insert(map, ro)
end

local mapframe = CreateFrame("Button", bg)
mapframe:SetHeight(32*22)
mapframe:SetWidth(32*22)
mapframe:SetBackgroundColor(0, 0, 0)

mapframe:SetPoint(nil, 0.5, UIParent, nil, 0.5)
mapframe:SetPoint("LEFT", UIParent, "LEFT", mapframe:GetTop(), nil)
mapframe:SetPoint("origin", 0, 0, mapframe, "TOPLEFT")
mapframe:SetPoint("origin", 32, 32, mapframe, "BOTTOMRIGHT")


-- Buffer layout:
-- X Y
-- T0 T1
-- R G B A
-- Conveniently this nicely fits in 32 bytes
local sampletex = Texture("sea")
local xadj = 1 / sampletex:GetWidth() / 2
local yadj = 1 / sampletex:GetHeight() / 2
local tesx = xadj
local tesy = yadj
local teex = sampletex:GetWidth() / sampletex:GetInternalWidth() + xadj
local teey = sampletex:GetHeight() / sampletex:GetInternalHeight() + yadj

local rend = {}
for i = 1, 32*32 do
  table.insert(rend, {})
end

-- Data is a key/value from image to alpha
local function SetSquare(x, y, data)
  local idx = x + 32 * (y - 1)
  rend[idx].x = x
  rend[idx].y = y
  rend[idx].data = data
end

local red_amount = 0

-- Refreshes the drawing, calls SetSquare
local function RefreshSquare(x, y)
  local images = {}
  local spot = map[x][y]
  local opac = 0
  if spot.colonized then
    opac = 1
  elseif spot.explored then
    opac = 0.5
  end
  
  if opac == 0 then
    -- hurf durf etc
    SetSquare(x, y, {})
    return
  end
  
  if not spot.tag then
    images.sea = opac
    images.sea_red = opac * red_amount
    if spot.civilized then
      images.boat = opac
    end
  else
    images[spot.tag] = opac
    if spot.civilized == "normal" then
      images.civilize = opac
    elseif spot.civilized == "lode" then
      images.lode = opac
    end
  end
  
  if spot.targeted and map_targeted_shown then
    images.targeted = 0.5
  end
  
  SetSquare(x, y, images)
end

local squares_buffer = GlBuffer()

local indices_reserve = {}
local indices_used = {}
local function indices_get()
  local buf = table.remove(indices_reserve)
  if not buf then
    buf = GlBuffer()
  end
  table.insert(indices_used, buf)
  return buf
end

-- do the actual drawing in here
function mapframe:Draw()
  table.append(indices_reserve, indices_used)
  indices_used = {}
  
  for x = 1, 32 do
    for y = 1, 32 do
      RefreshSquare(x, y)
    end
  end
  
  local draworder = {"sea", "sea_red", "field", "forest", "mountain", "civilize", "lode", "boat", "targeted"}
  
  local vertices = {}
  local indices = {}
  
  for _, k in ipairs(draworder) do
    indices[k] = {}
  end
  
  local iidx = 0
  for _, item in pairs(rend) do
    for tex, alph in pairs(item.data) do
      for i = 1, 4 do
        table.insert(indices[tex], iidx)
        iidx = iidx + 1
      end
      
      table.insert(vertices, item.x - 1)
      table.insert(vertices, item.y - 1)
      table.insert(vertices, tesx)
      table.insert(vertices, tesy)
      table.insert(vertices, 1)
      table.insert(vertices, 1)
      table.insert(vertices, 1)
      table.insert(vertices, alph)
      
      table.insert(vertices, item.x)
      table.insert(vertices, item.y - 1)
      table.insert(vertices, teex)
      table.insert(vertices, tesy)
      table.insert(vertices, 1)
      table.insert(vertices, 1)
      table.insert(vertices, 1)
      table.insert(vertices, alph)
      
      table.insert(vertices, item.x)
      table.insert(vertices, item.y)
      table.insert(vertices, teex)
      table.insert(vertices, teey)
      table.insert(vertices, 1)
      table.insert(vertices, 1)
      table.insert(vertices, 1)
      table.insert(vertices, alph)
      
      table.insert(vertices, item.x - 1)
      table.insert(vertices, item.y)
      table.insert(vertices, tesx)
      table.insert(vertices, teey)
      table.insert(vertices, 1)
      table.insert(vertices, 1)
      table.insert(vertices, 1)
      table.insert(vertices, alph)
    end
  end
  
  gl.EnableClientState("VERTEX_ARRAY")
  gl.EnableClientState("TEXTURE_COORD_ARRAY")
  gl.EnableClientState("COLOR_ARRAY")
  gl.BindBuffer("ARRAY_BUFFER", squares_buffer:get())
  gl.BufferData("ARRAY_BUFFER", "FLOAT", vertices, "STREAM_DRAW")
  
  gl.ClientActiveTexture(1)
  gl.VertexPointer(2, "FLOAT", 32, 0)
  gl.TexCoordPointer(2, "FLOAT", 32, 8)
  gl.ColorPointer(4, "FLOAT", 32, 16)
  
  for _, tex in ipairs(draworder) do
    Texture(tex):SetTexture()
    
    gl.BindBuffer("ELEMENT_ARRAY_BUFFER", indices_get():get())
    gl.BufferData("ELEMENT_ARRAY_BUFFER", "UNSIGNED_INT", indices[tex], "STREAM_DRAW")
    gl.DrawElements("QUADS", #indices[tex], "UNSIGNED_INT", 0)
  end
  
  gl.DisableClientState("VERTEX_ARRAY")
  gl.DisableClientState("TEXTURE_COORD_ARRAY")
  gl.DisableClientState("COLOR_ARRAY")
  
  gl.BindBuffer("ARRAY_BUFFER", 0)
  gl.BindBuffer("ELEMENT_ARRAY_BUFFER", 0)
end
function mapframe:Click()
  self:Update()
  if inside_x then
    event(inside_x, inside_y, true)
  end
end
function mapframe:Update()
  local ix, iy = self:GetMouse()
  ix, iy = math.floor(ix + 1), math.floor(iy + 1)
  if ix >= 1 and ix <= 32 and iy >= 1 and iy <= 32 then
    inside_x, inside_y = ix, iy
  else
    inside_x, inside_y = nil, nil
  end
end


local config_people_per_city = 4

boats, fields, mountains, forests, lodestones = 0, 0, 0, 0, 0
food, wood, metal, magic = 250, 250, 250, 250
magic_needed = 20000

local food_income_base = 20
local wood_income_base = 20
local metal_income_base = 20

local people_free -- kept as temporary indicators

local side_ui = CreateFrame("Frame", bg)

local some_bd = nil

local function make_backdrop(fram, r, g, b, ico, ttfunc)
  local bg = CreateFrame("Frame", bg)
  bg:SetLayer(-1)
  bg:SetBackgroundColor(r, g, b, 0.2)
  bg:SetPoint("TOPLEFT", fram, "TOPLEFT", -5, -5)
  bg:SetPoint("BOTTOMRIGHT", fram, "BOTTOMRIGHT", 5, 5)
  bg:SetPoint("RIGHT", UIParent, "RIGHT", -15, nil)
  
  if ico then
    local ti = CreateFrame("Texture", bg)
    ti:SetTexture(ico)
    ti:SetPoint(nil, 0.5, bg, nil, 0.5)
    ti:SetPoint(1, nil, bg, 1, nil, -(bg:GetHeight() - ti:GetHeight()) / 2)
  end
  
  if ttfunc then
    tooltippy(bg, ttfunc)
  end
  
  if not some_bd then some_bd = bg end
end

local indicator_sizes = 21

local people_indicator = CreateFrame("Text", bg)
people_indicator:SetText("People")
people_indicator:SetColor(1, 1, 1)
people_indicator:SetSize(indicator_sizes)
people_indicator:SetPoint("TOPLEFT", mapframe, "TOPRIGHT", 30, 0)
make_backdrop(people_indicator, 1, 1, 1, "civilize", function ()
  return ("People are the workers in your civilization. Each city provides %d people. Each city, boat, or lodestone consumes one person. If you run out of people, any excess structures are randomly destroyed and you cannot build more.\n\n%d cities: %d people produced\n%d structures: %d people consumed"):format(config_people_per_city, fields + mountains + forests, (fields + mountains + forests) * config_people_per_city, boats + fields + mountains + forests + lodestones, boats + fields + mountains + forests + lodestones)
end)

local food_indicator = CreateFrame("Text", bg)
food_indicator:SetText("Food")
food_indicator:SetColor(1, 1, 1)
food_indicator:SetSize(indicator_sizes)
food_indicator:SetPoint("TOPLEFT", people_indicator, "BOTTOMLEFT", 0, 10)
make_backdrop(food_indicator, 1, 1, 0, "field", function ()
  return ("Food is used to build cities, maintain cities and lodestones, and send scouting expeditions. Each city creates one food per turn and consumes one food per turn. Cities on yellow fields generate two extra food per turn. Lodestones consume one food per turn.\n\nRunning out of food is bad.\n\nIncome: %d cities + %d fields * 2 = %d\nExpenses: %d cities + %d lodestones = %d"):format(fields + mountains + forests, fields, fields + mountains + forests + fields * 2, fields + mountains + forests, lodestones, fields + mountains + forests + lodestones)
end)

local wood_indicator = CreateFrame("Text", bg)
wood_indicator:SetText("Wood")
wood_indicator:SetColor(1, 1, 1)
wood_indicator:SetSize(indicator_sizes)
wood_indicator:SetPoint("TOPLEFT", food_indicator, "BOTTOMLEFT", 0, 10)
make_backdrop(wood_indicator, 0, 1, 0, "forest", function ()
  return ("Wood is used to build cities, lodestones, and ships for travel. Each city creates one wood per turn. Cities on green forests generate two extra wood per turn.\n\nIncome: %d cities + %d forests * 2 = %d"):format(fields + mountains + forests, forests, fields + mountains + forests + forests * 2)
end)

local metal_indicator = CreateFrame("Text", bg)
metal_indicator:SetText("Metal")
metal_indicator:SetColor(1, 1, 1)
metal_indicator:SetSize(indicator_sizes)
metal_indicator:SetPoint("TOPLEFT", wood_indicator, "BOTTOMLEFT", 0, 10)
make_backdrop(metal_indicator, 1, 0.75, 1, "mountain", function ()
  return ("Metal is used to build cities and lodestones, as well as defend your civilization against monster attacks. Each city creates one metal per turn. Cities on auburn mountains generate two extra metal per turn.\n\nIncome: %d cities + %d mountains * 2 = %d"):format(fields + mountains + forests, mountains, fields + mountains + forests + mountains * 2)
end)

local magic_indicator = CreateFrame("Text", bg)
magic_indicator:SetText("Magic")
magic_indicator:SetColor(1, 1, 1)
magic_indicator:SetSize(indicator_sizes)
magic_indicator:SetPoint("TOPLEFT", metal_indicator, "BOTTOMLEFT", 0, 10)
make_backdrop(magic_indicator, 0, 0, 1, "lode", function ()
  return ("Magic is used to defend your civilization against monster attacks. More importantly, it is used to cast the spell Escape, your only way to escape this doomed planet, journey to a new world, and win the game. Each lodestone generates five magic per turn.\n\nIncome: %d lodestones * 5 = %d"):format(lodestones, lodestones * 5)
end)



local monster_indicator = CreateFrame("Text", side_ui)
monster_indicator:SetText("Next monster . . .")
monster_indicator:SetColor(1, 1, 1)
monster_indicator:SetSize(indicator_sizes)
monster_indicator:SetPoint("TOPLEFT", magic_indicator, "BOTTOMLEFT", 0, 10)
make_backdrop(monster_indicator, 0, 0, 0, nil, function ()
  return ("Every %d turns you will be attacked by a fearsome monster. The monster can only be defeated by the combined efforts of warriors and sorcerors, represented by metal and magic. Skimping on one will require far larger quantities of the other. Defeating the monster will provide a bounty of meat and bones, immediately transformed into food and \"wood\". Failing to defeat the monster will result in a large fraction of your buildings destroyed. Each monster will primarily target a randomly chosen structure type - sometimes it may not be worth the resources to defend!"):format(ticks_per_monster)
end)



local costs = {
  scout = {50, 0, 0, 0},
  travel = {0, 50, 0, 1},
  civilize = {50, 100, 50, 0},
  lodestone = {0, 50, 250, 1},
}

local function update_stats(freq)
  local rv = {}
  
  local cities = fields + mountains + forests
  rv.food_income = fields * 2 * freq + cities
  rv.food_outgo = fields + mountains + forests + lodestones
  rv.wood_income = forests * 2 * freq + cities
  rv.metal_income = mountains * 2 * freq + cities
  rv.people_total = (fields + mountains + forests) * config_people_per_city
  rv.people_needed = fields + mountains + forests + lodestones + boats
  rv.magic_income = lodestones * 5
  
  return rv
end
local function update_gui(freq)
  stats = update_stats(freq)
  
  people_indicator:SetText(("People: %d / %d"):format(stats.people_total, stats.people_needed))
  food_indicator:SetText(("Food: %d %+d"):format(food, stats.food_income - stats.food_outgo))
  wood_indicator:SetText(("Wood: %d +%d"):format(wood, stats.wood_income))
  metal_indicator:SetText(("Metal: %d +%d"):format(metal, stats.metal_income))
  magic_indicator:SetText(("Magic: %d +%d"):format(magic, stats.magic_income))
end

local monster_mode = false
local monster_ui = CreateFrame("Frame", bg)
monster_ui:Hide()

local descript_intro = CreateFrame("Text", monster_ui)
descript_intro:SetText("A monster assaults us.")
descript_intro:SetPoint("TOPLEFT", monster_indicator, "TOPLEFT")
descript_intro:SetColor(1, 1, 1)
descript_intro:SetSize(indicator_sizes)

local monster_name = CreateFrame("Text", monster_ui)
monster_name:SetPoint("TOPLEFT", monster_indicator, "TOPLEFT", 10, 60)
monster_name:SetSize(30)
monster_name:SetColor(1, 0.8, 0.8)

local monster_db = {"Corrupt Slime", "Bladed Dragon", "Wraith of the Lost", "Forgotten Beast", "Terror Swarm", "Warmaster", "Shadowlurker", "Dark Observer", "The Unstoppable", "He Who Sleeps"}

local monster_strength_tex = CreateFrame("Text", monster_ui)
monster_strength_tex:SetPoint("TOPLEFT", descript_intro, "TOPLEFT", 0, 120)
monster_strength_tex:SetText("Monster power:")
monster_strength_tex:SetSize(indicator_sizes)
monster_strength_tex:SetColor(1, 1, 1)

local defense_strength_tex = CreateFrame("Text", monster_ui)
defense_strength_tex:SetPoint("TOPLEFT", monster_strength_tex, "BOTTOMLEFT", 0, 10)
defense_strength_tex:SetText("Your defenses:")
defense_strength_tex:SetSize(indicator_sizes)
defense_strength_tex:SetColor(1, 1, 1)

local monster_strength = CreateFrame("Text", monster_ui)
monster_strength:SetPoint("TOP", monster_strength_tex, "TOP")
monster_strength:SetPoint("RIGHT", UIParent, "RIGHT", -20, nil)
monster_strength:SetSize(indicator_sizes)
monster_strength:SetColor(1, 0.8, 0.8)

local defense_strength = CreateFrame("Text", monster_ui)
defense_strength:SetPoint("TOP", defense_strength_tex, "TOP")
defense_strength:SetPoint("RIGHT", UIParent, "RIGHT", -20, nil)
defense_strength:SetSize(indicator_sizes)
defense_strength:SetColor(0.8, 0.8, 1)

local victory_butan = CreateFrame("Button", monster_ui)
local victory_text = CreateFrame("Text", victory_butan)
local victory_bg = CreateFrame("Frame", victory_butan)
victory_bg:SetPoint("LEFT", some_bd, "LEFT")
victory_bg:SetPoint("RIGHT", some_bd, "RIGHT")
victory_bg:SetPoint("TOP", defense_strength_tex, "TOP", nil, 220)
victory_bg:SetHeight(50)
victory_bg:SetBackgroundColor(0, 0, 0, 0.3)
victory_text:SetPoint("CENTER", victory_bg, "CENTER")
victory_text:SetText("Victory!")
victory_text:SetColor(1, 1, 1)
victory_text:SetLayer(1)
victory_text:SetSize(indicator_sizes)

local defeat_butan = CreateFrame("Button", monster_ui)
local defeat_text = CreateFrame("Text", defeat_butan)
local defeat_bg = CreateFrame("Frame", defeat_butan)
defeat_bg:SetPoint("LEFT", some_bd, "LEFT")
defeat_bg:SetPoint("RIGHT", some_bd, "RIGHT")
defeat_bg:SetPoint("TOP", defense_strength_tex, "TOP", nil, 280)
defeat_bg:SetHeight(50)
defeat_bg:SetBackgroundColor(0, 0, 0, 0.3)
defeat_text:SetPoint("CENTER", defeat_bg, "CENTER")
defeat_text:SetText("Retreat.")
defeat_text:SetColor(1, 1, 1)
defeat_text:SetLayer(1)
defeat_text:SetSize(indicator_sizes)


local result = {magic = 0, metal = 0}

tooltippy(victory_bg, function ()
  return ("Claim victory over the demon! If your defenses exceed its power, you can defeat it and transform its carcass into %d food and wood."):format(result.prizes or 0)
end)
tooltippy(defeat_bg, function ()
  return ("Abandon the sieged structures, leaving them to be destroyed by the monster. Every flashing structure will be immediately razed.")
end)

local widget_resync
local function make_widget(token, vert)
  local center = CreateFrame("Text", monster_ui)
  center:SetPoint(0.5, nil, some_bd, 0.5, nil)
  center:SetPoint(nil, 0.5, defense_strength_tex, nil, 0.5, nil, vert)
  center:SetColor(1, 1, 1)
  center:SetSize(indicator_sizes)
  
  local centbg = CreateFrame("Frame", monster_ui)
  centbg:SetLayer(-1)
  centbg:SetPoint("CENTER", center, "CENTER")
  centbg:SetHeight(center:GetHeight() + 10)
  centbg:SetWidth(140)
  centbg:SetBackgroundColor(0, 0, 0, 0.3)
  
  local mbutton = CreateFrame("Button", monster_ui)
  function mbutton:Click()
    assert(monster_mode)
    --if not monster_mode then return end -- what
    if result[token] - 10 < 0 then
    else
      result[token] = result[token] - 10
    end
    widget_resync()
  end
  mbutton.Repeat = mbutton.Click
  local pbutton = CreateFrame("Button", monster_ui)
  function pbutton:Click()
  if not monster_mode then return end -- what
    if result[token] + 10 > _G[token] then
    else
      result[token] = result[token] + 10
    end
    widget_resync()
  end
  pbutton.Repeat = pbutton.Click
  
  local mmbutton = CreateFrame("Button", monster_ui)
  function mmbutton:Click()
    assert(monster_mode)
    --if not monster_mode then return end -- what
    if result[token] - 1000 < 0 then
      result[token] = 0
    else
      result[token] = result[token] - 1000
    end
    widget_resync()
  end
  mmbutton.Repeat = mmbutton.Click
  local ppbutton = CreateFrame("Button", monster_ui)
  function ppbutton:Click()
  if not monster_mode then return end -- what
    if result[token] + 1000 > _G[token] then
      result[token] = math.floor(_G[token] / 10) * 10
    else
      result[token] = result[token] + 1000
    end
    widget_resync()
  end
  ppbutton.Repeat = ppbutton.Click
  
    
  local mm_bg = CreateFrame("Frame", mmbutton)
  mm_bg:SetPoint("TOP", centbg, "TOP")
  mm_bg:SetPoint("BOTTOM", centbg, "BOTTOM")
  mm_bg:SetPoint("LEFT", some_bd, "LEFT")
  mm_bg:SetWidth(15)
  mm_bg:SetLayer(-1)
  mm_bg:SetBackgroundColor(centbg:GetBackgroundColor())
  
  local pp_bg = CreateFrame("Frame", ppbutton)
  pp_bg:SetPoint("TOP", centbg, "TOP")
  pp_bg:SetPoint("BOTTOM", centbg, "BOTTOM")
  pp_bg:SetPoint("RIGHT", some_bd, "RIGHT")
  pp_bg:SetWidth(15)
  pp_bg:SetLayer(-1)
  pp_bg:SetBackgroundColor(centbg:GetBackgroundColor())
  
  local minus_bg = CreateFrame("Frame", mbutton)
  minus_bg:SetPoint("TOPRIGHT", centbg, "TOPLEFT", -5, 0)
  minus_bg:SetPoint("BOTTOM", centbg, "BOTTOM")
  minus_bg:SetPoint("LEFT", mm_bg, "RIGHT", 5, 0)
  minus_bg:SetLayer(-1)
  minus_bg:SetBackgroundColor(centbg:GetBackgroundColor())
  
  local plus_bg = CreateFrame("Frame", pbutton)
  plus_bg:SetPoint("TOPLEFT", centbg, "TOPRIGHT", 5, 0)
  plus_bg:SetPoint("BOTTOM", centbg, "BOTTOM")
  plus_bg:SetPoint("RIGHT", pp_bg, "LEFT", -5, 0)
  plus_bg:SetLayer(-1)
  plus_bg:SetBackgroundColor(centbg:GetBackgroundColor())

  
  centbg:SetBackgroundColor(0, 0, 0, 0.1)
  
  local minus_text = CreateFrame("Text", mbutton)
  minus_text:SetPoint("CENTER", minus_bg, "CENTER")
  minus_text:SetColor(1, 1, 1)
  minus_text:SetSize(15)
  minus_text:SetText("-")
  
  local plus_text = CreateFrame("Text", pbutton)
  plus_text:SetPoint("CENTER", plus_bg, "CENTER")
  plus_text:SetColor(1, 1, 1)
  plus_text:SetSize(15)
  plus_text:SetText("+")
  
  local mmt1 = CreateFrame("Text", mmbutton)
  mmt1:SetPoint("CENTER", mm_bg, 0.5, 1/3)
  mmt1:SetColor(1, 1, 1)
  mmt1:SetText("-")
  mmt1:SetSize(10)
  
  local mmt1 = CreateFrame("Text", mmbutton)
  mmt1:SetPoint("CENTER", mm_bg, 0.5, 2/3)
  mmt1:SetColor(1, 1, 1)
  mmt1:SetText("-")
  mmt1:SetSize(10)
  
  local ppt1 = CreateFrame("Text", ppbutton)
  ppt1:SetPoint("CENTER", pp_bg, 0.5, 1/3)
  ppt1:SetColor(1, 1, 1)
  ppt1:SetText("+")
  ppt1:SetSize(10)
  
  local ppt1 = CreateFrame("Text", ppbutton)
  ppt1:SetPoint("CENTER", pp_bg, 0.5, 2/3)
  ppt1:SetColor(1, 1, 1)
  ppt1:SetText("+")
  ppt1:SetSize(10)
  
  
  
  centbg:SetBackgroundColor(0, 0, 0, 0.1)
  
  return center
end

local metal_text = make_widget("metal", 100)
local magic_text = make_widget("magic", 140)

function widget_resync()
  metal_text:SetText("Metal: " .. result.metal)
  magic_text:SetText("Magic: " .. result.magic)
  
  result.defense = math.floor((100 + result.metal) * (100 + result.magic) / 1000)
  
  defense_strength:SetText(tostring(result.defense))
  
  if result.defense >= result.strength then
    victory_text:SetColor(1, 1, 1)
  else
    victory_text:SetColor(0.5, 0.5, 0.5)
  end
end

local levels = 4

ticks_per_monster = 35

local monster_per_fullred = 8

local monster_deaths
local monster_loop = coroutine.wrap(function ()
  local mstr = 60
  while true do
    for i = 1, ticks_per_monster do
      red_amount = red_amount + 1 / (ticks_per_monster * monster_per_fullred)
      red_amount = math.min(red_amount, 1)
      monster_indicator:SetText("Next monster: " .. ticks_per_monster - i + 1)
      
      if not game_ended then
        local fightmusicstart = ticks_per_monster / 2
        local fightmusicmax = 0.25
        params.gtable.music.fight:Gain(math.max(0, (i - fightmusicstart) / (ticks_per_monster - fightmusicstart) * fightmusicmax))
      end
      coroutine.yield()
    end
    
    bg:SetColor(1.0, 0.6, 0.3)
    
    monster_mode = true
    
    side_ui:Hide()
    monster_ui:Show()
    
    monster_name:SetText(monster_db[math.random(#monster_db)])
    monster_strength:SetText(tostring(mstr))
    
    result.magic = 0
    result.metal = 0
    --result.prizes = math.floor(math.sqrt(mstr) * 20)
    result.prizes = 300
    result.strength = mstr
    widget_resync()
    
    local weighting = {}
    local targ = math.random(5)
    weighting.boat = math.random() + 0.01 + ((targ == 1) and 2 or 0)
    weighting.lode = math.random() + 0.01 + ((targ == 2) and 2 or 0)
    weighting.field = math.random() + 0.01 + ((targ == 3) and 2 or 0)
    weighting.mountain = math.random() + 0.01 + ((targ == 4) and 2 or 0)
    weighting.forest = math.random() + 0.01 + ((targ == 5) and 2 or 0)
    
    
    
    local items = {}
    for x = 1, 32 do
      for y = 1, 32 do
        local weit = nil
        local mp = map[x][y]
        if mp.civilized then
          if mp.civilized == "lode" then
            weit = weighting.lode
          elseif not mp.tag then
            weit = weighting.boat
          elseif mp.tag == "forest" then
            weit = weighting.forest
          elseif mp.tag == "field" then
            weit = weighting.field
          elseif mp.tag == "mountain" then
            weit = weighting.mountain
          end
        end
        if weit then
          table.insert(items, {x, y, w = weit})
        end
      end
    end
    
    local doblast = math.ceil(#items / levels) + 2
    
    local targ = {}
    for i = 1, doblast do
      if #items == 0 then break end -- guess we're blowing it all up
      
      local weig = 0
      for _, v in ipairs(items) do
        weig = weig + v.w
      end
      
      weig = math.random() * weig
      
      for k, v in ipairs(items) do
        weig = weig - v.w
        if weig <= 0 then
          table.insert(targ, table.remove(items, k))
          break
        end
      end
    end
    
    monster_deaths = targ
    
    for _, v in ipairs(targ) do
      map[v[1]][v[2]].targeted = true
    end
    
    result.catastrophe = (#items == 0)
    
    params.gtable.music.fight:Gain(0.5)
    
    coroutine.yield()
    
    for _, v in ipairs(targ) do
      map[v[1]][v[2]].targeted = nil
    end
    
    side_ui:Show()
    monster_ui:Hide()
    
    monster_mode = false
    
    bg:SetColor(0.8, 0.8, 0.8)
    mstr = math.floor(mstr * 2)
  end
end)

function victory_butan:Click()
  if result.defense >= result.strength then
    monster_loop()
    metal = metal - result.metal
    magic = magic - result.magic
    food = food + result.prizes
    wood = wood + result.prizes
    update_gui(1)
    PlaySound("monster_boom")
  end
end
function defeat_butan:Click()
  monster_loop()
  for _, v in pairs(monster_deaths) do
    decivilize(v[1], v[2])
  end
  update_gui(1)
  if result.catastrophe then
    endgame("Defeat.")
    params.gtable.music.fight:Gain(0.5)
  end
  levels = levels - 1
  if levels < 1 then levels = 1 end
  PlaySound("monster_fail")
end

local function update_gui_and_tick(freq)
  freq = freq or 1

  local stats = update_stats(freq)
  
  food = food + stats.food_income - stats.food_outgo
  wood = wood + stats.wood_income
  metal = metal + stats.metal_income
  magic = magic + stats.magic_income
  
  people_free = stats.people_total - stats.people_needed
  
  -- deal with not having enough people/food here, trash stuff
  
  update_gui(freq)
  monster_loop()
end


local active_butan = nil
local semiactive_butan = nil

local butanz = {}
local butan_ankor = monster_indicator
local function make_butan(id, text, tooltip, func, mousein, mouseout)
  local butan = CreateFrame("Button", side_ui)
  local around = CreateFrame("Frame", butan)
  around:SetPoint("TOPLEFT", butan_ankor, "BOTTOMLEFT", 0, 20)
  around:SetPoint("RIGHT", UIParent, "RIGHT", -20, 0)
  around:SetHeight(50)
  around:SetBackgroundColor(0, 0, 0)
  local center = CreateFrame("Frame", around)
  center:SetPoint("TOPLEFT", around, "TOPLEFT", 10, 10)
  center:SetPoint("BOTTOMRIGHT", around, "BOTTOMRIGHT", -10, -10)
  center:SetBackgroundColor(0, 0, 0)
  local tex = CreateFrame("Text", center)
  tex:SetText(text)
  tex:SetColor(1, 1, 1)
  tex:SetSize(indicator_sizes)
  tex:SetPoint("CENTER", around, "CENTER")
  butan_ankor = around
  
  if func then
    butan.Click = func
  else
    function butan:Click()
      if game_ended then return end
      if active_butan == id then
        around:SetBackgroundColor(0, 0, 0)
        active_butan = nil
      else
        for _, v in pairs(butanz) do
          v:SetBackgroundColor(0, 0, 0)
        end
        
        around:SetBackgroundColor(1, 1, 1)
        active_butan = id
      end
    end
  end
  
  around.MouseIn = mousein
  around.MouseOut = mouseout
  
  tooltippy(around, tooltip)
  
  butanz[id] = around
end
local function minifresh_butans(semiactive)
  semiactive_butan = semiactive
  if not active_butan then
    for _, v in pairs(butanz) do
      v:SetBackgroundColor(0, 0, 0)
    end
    
    if semiactive and butanz[semiactive] then
      butanz[semiactive]:SetBackgroundColor(0.5, 0.5, 0.5)
    end
  end
end

make_butan("scout", "Scout", function ()
  return ("Scouting is your primary method of searching the world for more islands. Scouting consumes %d food and uncovers a 5x5 square centered around the target.\n\nThere are a total of 13 islands in the world."):format(costs.scout[1])
end)
make_butan("travel", "Travel", function ()
  return ("Traveling expands your empire via a network of ships. Each ship costs %d wood and one crewmember, and allows you travel to every adjacent square. Destroying the ship after construction will not reduce the bounds of your empire, but will free the crewmember to work on other jobs."):format(costs.travel[2])
end)
make_butan("civilize", "Civilize", function ()
  return ("Cities can be built for a fee of %d food, %d wood, and %d metal. Cities provide three people, one of which must staff the city. Cities produce one food, wood, and metal per turn, as well as a two-unit bonus depending on what type of square they're placed on. Cities consume one food per turn."):format(costs.civilize[1], costs.civilize[2], costs.civilize[3])
end)
make_butan("raze", "Raze", function ()
  return ("You may raze any structure you choose for free. Doing so does not reclaim any of the construction costs, but does stop any ongoing maintenance costs.")
end)
make_butan("work", "Work", function ()
  return ("Instead of building, you may spend a turn encouraging your citizens to produce more heavily. Each city will produce an extra two units of resources based on the tile they've been placed on. This does not affect lodestones. The resource chart in the top-right of the screen is currently showing you the results of ordering Work for a turn.")
end, function ()
  if game_ended then return end
  update_gui_and_tick(2)
end, function ()
  if game_ended then return end
  update_gui(2)
end, function ()
  update_gui(1)
end)
make_butan("lodestone", "Forge", function ()
  return ("Lodestones can be forged on any open land square for %d wood and %d metal. Lodestones require one person to staff and one food per turn to maintain. Lodestones produce five magic per turn."):format(costs.lodestone[2], costs.lodestone[3])
end)
make_butan("escape", "Escape", function ()
  return ("The Escape spell is your only path to victory. It will immediately teleport your entire civilization, no matter how large, off this doomed planet and to a safe haven. The Escape spell costs %d magic to cast."):format(magic_needed)
end, function ()
  if game_ended then return end
  if magic > magic_needed then
    for x = 1, 32 do
      for y = 1, 32 do
        if map[x][y].civilized then
          decivilize(x, y)  -- yay
        end
      end
    end
    PlaySound("teleport")
    endgame("VICTORY!")
    params.gtable.music.fight:Gain(0)
  end
end)

local function makedirs(dist, hole)
  local dx, dy = {}, {}
  for x = -dist, dist do
    for y = -dist, dist do
      if not hole or x ~= 0 or y ~= 0 then
        table.insert(dx, x)
        table.insert(dy, y)
      end
    end
  end
  return dx, dy
end
local cdx, cdy = makedirs(1, true)
local function inbounds(x, y)
  return x > 0 and y > 0 and x < 33 and y < 33
end
local function colonize(x, y)
  if inbounds(x, y) then
    if not map[x][y].colonized then
      map[x][y].colonized = true
      cached = nil
      if map[x][y].tag then
        for i = 1, #cdx do
          colonize(x + cdx[i], y + cdy[i])
        end
      end
    end
  end
end

local function civilize(x, y, lode)
  assert(not map[x][y].civilized)
  assert(not lode or map[x][y].tag)
  
  map[x][y].civilized = lode and "lode" or "normal"
  
  if lode then
    lodestones = lodestones + 1
  elseif map[x][y].tag == nil then
    boats = boats + 1
  elseif map[x][y].tag == "field" then
    fields = fields + 1
  elseif map[x][y].tag == "mountain" then
    mountains = mountains + 1
  elseif map[x][y].tag == "forest" then
    forests = forests + 1
  end
  
  cached = nil
  
  colonize(x, y)
  for i = 1, #cdx do
    colonize(x + cdx[i], y + cdy[i])
  end
end
function decivilize(x, y)
  local mp = map[x][y]
  
  assert(map[x][y].civilized)
  
  if map[x][y].civilized == "lode" then
    lodestones = lodestones - 1
  elseif map[x][y].tag == nil then
    boats = boats - 1
  elseif map[x][y].tag == "field" then
    fields = fields - 1
  elseif map[x][y].tag == "mountain" then
    mountains = mountains - 1
  elseif map[x][y].tag == "forest" then
    forests = forests - 1
  end
  
  map[x][y].civilized = nil
  
  cached = nil
end

local function expose(x, y)
  local dx, dy = makedirs(2, false)
  for i = 1, #dx do
    local tx, ty = x + dx[i], y + dy[i]
    if inbounds(tx, ty) and not map[tx][ty].colonized then
      map[tx][ty].explored = true
      cached = nil
    end
  end
end



local function event_has_resources(typ)
  if not costs[typ] then return true end
  return costs[typ][1] <= food and costs[typ][2] <= wood and costs[typ][3] <= metal and people_free >= costs[typ][4]
end
local function event_consume_resources(typ)
  if not costs[typ] then return end
  food = food - costs[typ][1]
  wood = wood - costs[typ][2]
  metal = metal - costs[typ][3]
end
local function event_resources_readable(typ)
  if not costs[typ] then return "" end
  local st = ""
  if costs[typ][1] > 0 then
    if #st > 0 then st = st .. " " end
    st = st .. costs[typ][1] .. " food"
  end
  if costs[typ][2] > 0 then
    if #st > 0 then st = st .. " " end
    st = st .. costs[typ][2] .. " wood"
  end
  if costs[typ][3] > 0 then
    if #st > 0 then st = st .. " " end
    st = st .. costs[typ][3] .. " metal"
  end
  if costs[typ][4] > 0 then
    if #st > 0 then st = st .. " " end
    st = st .. costs[typ][4] .. " people"
  end
  if #st > 0 then return "(" .. st .. ")" end
  return st
end
function event(x, y, actuallydo)
  if monster_mode then return "" end  -- disabled
  if game_ended then return "" end
  
  local inst = active_butan or semiactive_butan
  local mt = map[x][y]
  
  local didit = nil
  if inst == "scout" then
    if not event_has_resources("scout") then
      return "You don't have the resources needed to scout! " .. event_resources_readable("scout")
    end
    
    if actuallydo then
      PlaySound("scout")
      event_consume_resources("scout")
      expose(x, y)
    end

    didit = "Click to scout unexplored territory  " .. event_resources_readable("scout")
  elseif inst == "travel" then
    if not mt.tag and not mt.civilized and mt.colonized then
      if not event_has_resources("travel") then
        return "You don't have the resources needed to travel! " .. event_resources_readable("travel")
      end
      
      if actuallydo then
        PlaySound("boat")
        event_consume_resources("travel")
        civilize(x, y)
      end
      
      didit = "Click to construct a boat and expand your civilization's reach " .. event_resources_readable("travel")
    else
      return "You can only build a boat on water next to a discovered island or another boat"
    end
  elseif inst == "civilize" then
    if mt.tag and not mt.civilized and mt.colonized then
      if not event_has_resources("civilize") then
        return "You don't have the resources needed to civilize! " .. event_resources_readable("civilize")
      end
      
      if actuallydo then
        PlaySound("construct")
        event_consume_resources("civilize")
        civilize(x, y)
      end
      
      didit = "Click to construct a city and increase your civilization's wealth " .. event_resources_readable("civilize")
    else
      return "You can only build a city on an unoccupied square of a discovered island"
    end
  elseif inst == "raze" then
    if mt.civilized then
      if actuallydo then
        PlaySound("raze")
        decivilize(x, y)
      end
      
      didit = "Click to destroy a structure"
    else
      return "You can't destroy a structure that doesn't exist"
    end
  elseif inst == "work" then
    -- this is going to have to work differently anyway
  elseif inst == "lodestone" then
    if mt.tag and not mt.civilized and mt.colonized then
      if not event_has_resources("lodestone") then
        return "You don't have the resources needed to forge a lodestone! " .. event_resources_readable("lodestone")
      end
      
      if actuallydo then
        PlaySound("lode")
        event_consume_resources("lodestone")
        civilize(x, y, true)
      end
      
      didit = "Click to forge a lodestone " .. event_resources_readable("lodestone")
    else
      return "You can only build a lodestone on an unoccupied square of a discovered island"
    end
  end
  
  if didit and actuallydo then
    update_gui_and_tick()
  end
  
  return didit
end

do
  local islands = {}

  local dx, dy = {}, {}
  for tx = -2, 2 do
    for ty = -2, 2 do
      if tx ~= 0 or ty ~= 0 then
        table.insert(dx, tx)
        table.insert(dy, ty)
      end
    end
  end
  local ldx, ldy = {1, 0, 0, -1}, {0, 1, -1, 0}
  
  local function adjacentize(adj, x, y)
    for i = 1, #dx do
      local tx, ty = x + dx[i], y + dy[i]
      if inbounds(tx, ty) then
        if map[tx][ty].used then
          return true -- whuups
        end
      end
    end
    for i = 1, #ldx do
      local tx, ty = x + ldx[i], y + ldy[i]
      if inbounds(tx, ty) then
        if not map[tx][ty].tadjacent and not map[tx][ty].tused then
          table.insert(adj, {tx, ty})
          map[tx][ty].tadjacent = true
        end
      end
    end
  end
  local function cleanup(used, adj)
    for _, v in ipairs(used) do
      map[v[1]][v[2]].tused = nil
      map[v[1]][v[2]].tadjacent = nil
    end
    for _, v in ipairs(adj) do
      map[v[1]][v[2]].tused = nil
      map[v[1]][v[2]].tadjacent = nil
    end
  end
  
  local function makeisland(first)
    -- first choose a coordinate
    local x, y = math.random(32), math.random(32)
    if map[x][y].owned then return false end
    
    -- now we want to keep a list of taken squares, and a list of adjacent squares
    local used = {}
    local adj = {}
    
    map[x][y].tused = true
    
    table.insert(used, {x, y})
    if adjacentize(adj, x, y) then
      cleanup(used, adj)
      return false
    end
    
    local size
    if first then
      size = 6
    else
      size = math.random(5, 7)
    end
    size = size - 1
    
    for i = 1, size do
      local spot = math.random(#adj)
      local lx, ly = unpack(table.remove(adj, spot))
      if adjacentize(adj, lx, ly) then
        cleanup(used, adj)
        return false
      end
      map[lx][ly].tused = true
      table.insert(used, {lx, ly})
    end
    
    for _, v in pairs(used) do
      local x, y = unpack(v)
      map[x][y].used = true
    end
    
    table.insert(islands, used)
    
    return true
  end
  
  for i = 1, 13 do
    while true do
      if makeisland(i == 1) then break end
    end
  end
  
  local tociv = {}
  
  local function assign(label, spot, found)
    local x, y = unpack(spot)
    map[x][y].tag = label
    
    if found then
      table.insert(tociv, {x, y})
    end
  end
  
  local function portionisland(island, forest, field, mountain, foundfirst)
    if not forest then forest = #island - field - mountain end
    if not field then field = #island - forest - mountain end
    if not mountain then mountain = #island - field - forest end
    
    assert(#island == forest + field + mountain)
    
    for i = 1, forest do
      assign("forest", table.remove(island, math.random(#island)), foundfirst and i == 1)
    end
    for i = 1, field do
      assign("field", table.remove(island, math.random(#island)), foundfirst and i == 1)
    end
    for i = 1, mountain do
      assign("mountain", table.remove(island, math.random(#island)), foundfirst and i == 1)
    end
    assert(#island == 0)
  end
  
  portionisland(islands[1], 2, 2, 2, true)
  
  portionisland(islands[2], 1, 1, nil)
  portionisland(islands[3], 1, 1, nil)
  portionisland(islands[4], 1, 1, nil)
  portionisland(islands[5], 1, 1, nil)
  
  portionisland(islands[6], 1, nil, 1)
  portionisland(islands[7], 1, nil, 1)
  portionisland(islands[8], 1, nil, 1)
  portionisland(islands[9], 1, nil, 1)
  
  portionisland(islands[10], nil, 1, 1)
  portionisland(islands[11], nil, 1, 1)
  portionisland(islands[12], nil, 1, 1)
  portionisland(islands[13], nil, 1, 1)
  
  for _, v in pairs(tociv) do
    civilize(unpack(v))
  end
end

update_gui_and_tick()

function loop(...)
  if endgame_freeze then return end
  
  if math.mod(math.floor(os.time(), 2), 2) == 0 then
    map_targeted_shown = true
  else
    map_targeted_shown = false
  end

  local msg = ""
  
  if inside_x and inside_y then
    local mp = map[inside_x][inside_y]
    if not mp.colonized then
      minifresh_butans("scout")
    elseif not mp.tag and not mp.civilized then
      minifresh_butans("travel")
    elseif not mp.civilized then
      minifresh_butans("civilize")
    else
      minifresh_butans(nil)
    end
    
    msg = event(inside_x, inside_y) or ""
  else
    minifresh_butans(nil)
  end
  
  status:SetText(msg)
end
-- think there's more